#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <pthread.h>

#include <linux/slab.h>
#include <linux/radix-tree.h>
#include <linux/rcupdate.h>

#include "test.h"

static RADIX_TREE(tree, GFP_KERNEL);
static pthread_mutex_t tree_lock = PTHREAD_MUTEX_INITIALIZER;

#define item_insert(tree, index)	\
do {					\
	item_insert(tree, index);	\
	item_tag_set(tree, index, 0);	\
} while (0)

static void *simple_updater(void *arg)
{
	int i;

	for (i = 0; i < 1000; i++) {
		int j;

		for (j = 0; j < 1000; j++)
			item_insert(&tree, j);

		for (j = 0; j < 1000; j++)
			item_delete(&tree, (j+i)%1000);
//			item_delete(&tree, j);
	}

	printk("after simple_updater %d allocated\n", nr_allocated);

	return NULL;
}

static void *random_updater(void *arg)
{
	int i;

	for (i = 0; i < 1000; i++) {
		int j;
		for (j = 0; j < 1000; j++) {
			unsigned long idx = random() % 1000;
			item_insert(&tree, idx);
		}

		for (j = 0; j < 1000; j++) {
			unsigned long idx = random() % 1000;
			item_delete(&tree, idx);
		}
	}
	item_kill_tree(&tree);
	printk("after random_updater %d allocated\n", nr_allocated);

	return NULL;
}

static void *height_change_updater(void *arg)
{
	int i;

	for (i = 0; i < 100000; i++) {
		item_insert(&tree, 0);
		item_delete(&tree, 0);

		item_insert(&tree, 1000000);
		item_insert(&tree, 1);
		item_insert(&tree, 2);
		item_insert(&tree, 3);
		item_insert(&tree, 10000);
		item_delete(&tree, 1000000);
		item_insert(&tree, 5);
		item_insert(&tree, 6);
		item_insert(&tree, 7);
		item_delete(&tree, 2);
		item_delete(&tree, 6);
		item_delete(&tree, 3);
		item_delete(&tree, 10000);
		item_delete(&tree, 1);
		item_delete(&tree, 5);
		item_delete(&tree, 7);
	}
	printk("after height_change_updater %d allocated\n", nr_allocated);

	return NULL;
}

static void *indirect_updater(void *arg)
{
	int i;

	for (i = 0; i < 1000000; i++) {
		item_insert(&tree, 1);
		item_delete(&tree, 1);
	}
	printk("after indirect_updater %d allocated\n", nr_allocated);

	return NULL;
}

static void *direct_updater(void *arg)
{
	int i;

	for (i = 0; i < 1000000; i++) {
		item_insert(&tree, 0);
		item_delete(&tree, 0);
	}
	printk("after direct_updater %d allocated\n", nr_allocated);

	return NULL;
}

static void *direct_change_updater(void *arg)
{
	int i;

	for (i = 0; i < 100000; i++) {
		item_insert(&tree, 1);
		item_delete(&tree, 1);

		item_insert(&tree, 0);
		item_insert(&tree, 1);
		item_delete(&tree, 1);
		item_delete(&tree, 0);

		item_insert(&tree, 0);
		item_insert(&tree, 1);
		item_delete(&tree, 0);
		item_delete(&tree, 1);

		item_insert(&tree, 1);
		item_insert(&tree, 0);
		item_delete(&tree, 1);
		item_delete(&tree, 0);
	}
	printk("after direct_change_updater %d allocated\n", nr_allocated);

	return NULL;
}

static int run_readers = 1;

static void *simple_reader(void *arg)
{
	int lookup = 0, found = 0;

	while (run_readers) {
		int j;
		rcu_read_lock();
		for (j = 0; j < 1000; j++) {
			struct item *item = radix_tree_lookup(&tree, j);
			if (item) {
				found++;
				assert(item->index == j);
			}
			lookup++;
		}
		rcu_read_unlock();
	}

	printk("simple_reader found %d out of %d\n", found, lookup);

	return NULL;
}

static void *simple_gang_reader(void *arg)
{
	int lookup = 0, found = 0;

	while (run_readers) {
		int j;
		struct item *items[32];
		rcu_read_lock();

		for (j = 0; j < 1000;) {
			int k;
			int nr;
			nr = radix_tree_gang_lookup_tag(&tree, (void **)items, j, 32, 0);
			
			for (k = 0; k < nr; k++) {
				assert(items[k]->index >= j);
				j = items[k]->index+1;
			}
			found += nr;
			lookup++;
			if (!nr)
				break;
		}
		rcu_read_unlock();
	}

	printk("simple_reader found %d out of %d\n", found, lookup);

	return NULL;
}

static void *direct_reader(void *arg)
{
	int lookup = 0, found = 0;

	while (run_readers) {
		int j;
		rcu_read_lock();
		for (j = 0; j < 1000; j++) {
			struct item *item;
			item = radix_tree_lookup(&tree, 0);
			if (item) {
				found++;
				assert(item->index == 0);
			}
			lookup++;
		}
		rcu_read_unlock();
	}

	printk("direct_reader found %d out of %d\n", found, lookup);

	return NULL;
}

static void *zeroone_reader(void *arg)
{
	int lookup = 0, found = 0;

	while (run_readers) {
		int j;
		rcu_read_lock();
		for (j = 0; j < 1000; j++) {
			struct item *item = radix_tree_lookup(&tree, j&1);
			if (item) {
				found++;
				assert(item->index == (j&1));
			}
			lookup++;
		}
		rcu_read_unlock();
	}

	printk("zeroone_reader found %d out of %d\n", found, lookup);

	return NULL;
}

int main()
{
	int i, t = 0;
	pthread_t readers[4];

	srandom(10);

	rcu_init();
	radix_tree_init();

	pthread_create(&readers[t++], NULL, &simple_reader, NULL);
	pthread_create(&readers[t++], NULL, &simple_gang_reader, NULL);
	pthread_create(&readers[t++], NULL, &direct_reader, NULL);
	pthread_create(&readers[t++], NULL, &zeroone_reader, NULL);

	simple_updater(NULL);
	random_updater(NULL);
	height_change_updater(NULL);
	direct_change_updater(NULL);
	direct_updater(NULL);
	indirect_updater(NULL);

	run_readers = 0;

	for (i = 0; i < t; i++)
		pthread_join(readers[i], NULL);
	printk("after readers finish %d allocated\n", nr_allocated);

	item_kill_tree(&tree);
	printk("after tree killed %d allocated\n", nr_allocated);

	rcu_exit();

	exit(0);
}
